from collections import Counter
import enaml
from enaml.qt.qt_application import QtApplication

import argparse
import os.path

from abr.parsers import Parser

from . import add_default_arguments, parse_args


with enaml.imports():
    from abr.main_window import SerialWindow
    from abr.presenter import SerialWaveformPresenter


def main():
    parser = argparse.ArgumentParser("abr_loop")
    add_default_arguments(parser)
    parser.add_argument('dirnames', nargs='*')
    parser.add_argument('--list', action='store_true')
    parser.add_argument('--skip-errors', action='store_true')
    options = parse_args(parser)
    parser = options['parser']

    unprocessed = []
    for dirname in options['dirnames']:
        files = parser.find_unprocessed(dirname)
        unprocessed.extend(files)

    if len(unprocessed) == 0:
        print('No files to process')
    elif options['list']:
        counts = Counter(f for f, _ in unprocessed)
        for filename, n in counts.items():
            filename = os.path.basename(filename)
            print(f'{filename} ({n})')
    else:
        app = QtApplication()
        presenter = SerialWaveformPresenter(parser=parser,
                                            latencies=options['latencies'],
                                            unprocessed=unprocessed)
        view = SerialWindow(presenter=presenter)
        view.show()
        app.start()
        app.stop()



import argparse

from atom.api import Bool, Typed, Unicode

import enaml
from enaml.core.api import d_, Declarative
from enaml.qt.qt_application import QtApplication
with enaml.imports():
    from abr.main_window import CompareWindow

from matplotlib import pylab as pl
import numpy as np
import pandas as pd

from . import add_default_arguments, parse_args


class Compare(Declarative):

    data = Typed(pd.DataFrame)
    x_column = d_(Unicode())
    y_column = d_(Unicode())
    as_difference = d_(Bool(True))
    jitter = d_(Bool(True))
    axes = Typed(pl.Axes)
    figure = Typed(pl.Figure)
    selected = Typed(list)

    def _default_figure(self):
        return pl.Figure()

    def _default_axes(self):
        return self.figure.add_subplot(111)

    def _observe_data(self, event):
        self._update_plot()

    def _observe_x_column(self, event):
        self._update_plot()

    def _observe_y_column(self, event):
        self._update_plot()

    def _observe_as_difference(self, event):
        self._update_plot()

    def _observe_jitter(self, event):
        self._update_plot()

    def _default_x_column(self):
        return self.data.columns[0]

    def _default_y_column(self):
        i = 1 if (len(self.data.columns) > 1) else 0
        return self.data.columns[i]

    def _update_plot(self):
        x = self.data[self.x_column].copy()
        y = self.data[self.y_column].copy()
        if self.as_difference:
            y -= x
        if self.jitter:
            x += np.random.uniform(-1, 1, len(x))
            y += np.random.uniform(-1, 1, len(x))

        self.axes.clear()
        self.axes.plot(x, y, 'ko', picker=4, mec='w', mew=1)
        if self.figure.canvas is not None:
            self.figure.canvas.draw()

    def pick_handler(self, event):
        rows = self.data.iloc[event.ind]
        files = list(rows.index.get_level_values('raw_file'))
        frequencies = list(rows.index.get_level_values('frequency'))
        self.selected = list(zip(files, frequencies))


def main():
    parser = argparse.ArgumentParser("abr_compare")
    add_default_arguments(parser, waves=False)
    parser.add_argument('directory')
    options = parse_args(parser, waves=False)

    data = options['parser'].load_analyses(options['directory'])
    data = data.reset_index(['analyzed_file'], drop=True).unstack('user')
    data = data.sort_index()

    figure, axes = pl.subplots(1, 1)
    compare = Compare(data=data)

    app = QtApplication()
    view = CompareWindow(parser=options['parser'], compare=compare)
    view.show()
    app.start()
    app.stop()


if __name__ == '__main__':
    main()



from pathlib import Path

import matplotlib as mp
from matplotlib.pylab import setp

rc_path = Path(__file__).parent / 'matplotlibrc'
mp.rc_file(str(rc_path))


class StylePlot:

    HIDDEN = {'alpha': 0}

    def update(self):
        self.update_plot()
        setp(self.plot, **self.get_style())

    def get_style(self):
        raise NotImplementedError

    def update_plot(self):
        pass


class PointPlot(StylePlot):

    PEAK = {
        'linestyle':        ' ',
        'marker':           'o',
        'zorder':           20,
        'alpha':            1,
        'markersize':       8,
        'markeredgewidth':  1,
        'markeredgecolor':  (0, 0, 0)
    }

    PEAK_FADED = {
        'linestyle':        ' ',
        'marker':           'o',
        'zorder':           20,
        'alpha':            0.5,
        'markersize':       8,
        'markeredgewidth':  1,
        'markeredgecolor':  (0, 0, 0)
    }

    VALLEY = {
        'linestyle':        ' ',
        'marker':           '^',
        'zorder':           20,
        'alpha':            1,
        'markersize':       9,
        'markeredgewidth':  1,
        'markeredgecolor':  (0, 0, 0)
    }

    TOGGLE = {
        'linestyle':        ' ',
        'marker':           's',
        'zorder':           100,
        'alpha':            1,
        'markersize':       8,
        'markeredgewidth':  1,
        'markerfacecolor':  (1, 1, 1),
        'markeredgecolor':  (0, 0, 0)
    }

    COLORS = [(1, 0, 0), (1, 1, 0), (0, 1, 0), (0, 1, 1), (0, 0, 1)]

    def __init__(self, parent, figure, point):
        self.figure = figure
        self.parent = parent
        self.point = point
        self.plot, = self.figure.plot(0, 0, transform=parent.transform,
                                      clip_on=False, picker=10)
        self.current = False
        self.update()

    def get_style(self):
        # Hide subthreshold points
        if self.parent.waveform.is_subthreshold():
            return self.HIDDEN

        # Return toggled value
        if self.current and self.parent.current:
            return self.TOGGLE

        # Fallback to this
        style = self.PEAK if self.point.is_peak() else self.VALLEY
        index = self.point.wave_number-1
        c = self.COLORS[self.point.wave_number-1]
        style['c'] = c
        style['markerfacecolor'] = c
        return style

    def update_plot(self):
        self.plot.set_data(self.point.x, self.point.y)

    def remove(self):
        self.plot.remove()


class WaveformPlot(StylePlot):

    CUR_PLOT = {
        'c':            (0, 0, 0),
        'linewidth':    4,
        'linestyle':    '-',
        'zorder':       20,
    }
    PLOT = {
        'c':            (0, 0, 0),
        'linewidth':    2,
        'linestyle':    '-',
        'zorder':       10,
    }
    CUR_SUBTH_PLOT = {
        'c':            (0.75, 0.75, 0.75),
        'linewidth':    4,
        'linestyle':    '-',
        'zorder':       10,
    }
    SUBTH_PLOT = {
        'c':            (0.75, 0.75, 0.75),
        'linewidth':    2,
        'linestyle':    '-',
        'zorder':       10,
    }

    def __init__(self, waveform, axis, transform):
        self.axis = axis
        self.waveform = waveform
        self.current = False
        self.point_plots = {}
        self.transform = transform

        # Create the plot
        self.plot, = self.axis.plot(self.waveform.x, self.waveform.y, 'k-',
                                    transform=transform, clip_on=False,
                                    picker=10)
        self.update()

    STYLE = {
        (True,  True):  CUR_PLOT,
        (True,  False): CUR_SUBTH_PLOT,
        (False, True):  PLOT,
        (False, False): SUBTH_PLOT,
    }

    def get_style(self):
        style = self.current, self.waveform.is_suprathreshold()
        return self.STYLE[style]

    def update(self):
        # Check to see if new points were added (e.g. valleys)
        for key, point in self.waveform.points.items():
            if key not in self.point_plots:
                self.point_plots[key] = PointPlot(self, self.axis, point)

        for key, point_plot in list(self.point_plots.items()):
            point = self.waveform.points.get(key)
            if point is None:
                point_plot.remove()
                del self.point_plots[key]
            elif point != point_plot.point:
                point_plot.point = self.waveform.points[key]

        for p in self.point_plots.values():
            p.update()

        super().update()



import glob
import os.path
from pathlib import Path

import numpy as np
import pandas as pd

from abr.datatype import ABRWaveform, ABRSeries


base_template = 'ABR -1.0ms to 9.0ms{}average waveforms.csv'
nofilter_template = base_template.format(' ')
filter_template = base_template.format(' with {:.0f}Hz to {:.0f}Hz filter ')


def get_filename(pathname, filter_settings):
    if filter_settings is not None:
        filename = filter_template.format(
            filter_settings['highpass'],
            filter_settings['lowpass'])
    else:
        filename = nofilter_template

    if pathname.name == filename:
        return pathname
    else:
        return pathname / filename


def load(base_directory, filter_settings=None, frequencies=None):
    filename = get_filename(base_directory, filter_settings)
    if frequencies is not None and np.isscalar(frequencies):
        frequencies = [frequencies]

    data = pd.io.parsers.read_csv(filename, header=[0, 1], index_col=0).T
    fs = np.mean(np.diff(data.columns.values)**-1)
    data.columns *= 1e3
    waveforms = {}
    for (frequency, level), w in data.iterrows():
        frequency = float(frequency)
        level = float(level)
        if frequencies is not None:
            if frequency not in frequencies:
                continue
        frequency = float(frequency)
        level = float(level)
        waveform = ABRWaveform(fs, w, level)
        waveforms.setdefault(frequency, []).append(waveform)

    series = []
    for frequency, stack in waveforms.items():
        s = ABRSeries(stack, frequency)
        s.filename = filename
        s.id = filename.parent.name
        series.append(s)

    return series


def get_frequencies(filename, filter_settings):
    data = pd.io.parsers.read_csv(filename, header=[0, 1], index_col=0).T
    frequencies = np.unique(data.index.get_level_values('frequency'))
    return frequencies.astype('float')


def find_all(dirname, filter_settings):
    results = []
    for pathname in Path(dirname).glob('*abr*'):
        if pathname.is_dir() :
            filename = get_filename(pathname, filter_settings)
            for frequency in get_frequencies(filename, filter_settings):
                results.append((filename, frequency))
    return results



import argparse
from pathlib import Path

import enaml
from enaml.application import deferred_call
from enaml.qt.qt_application import QtApplication

from . import add_default_arguments, parse_args
from abr.parsers import Parser

with enaml.imports():
    from abr.main_window import DNDWindow, load_files
    from abr.presenter import WaveformPresenter


def main():
    parser = argparse.ArgumentParser('abr_gui')
    add_default_arguments(parser)
    parser.add_argument('--demo', action='store_true', dest='demo',
                        default=False, help='Load demo data')
    parser.add_argument('filenames', nargs='*')
    options = parse_args(parser)

    app = QtApplication()
    view = DNDWindow(parser=options['parser'], latencies=options['latencies'])

    filenames = [(Path(f), None) for f in options['filenames']]

    deferred_call(load_files, options['parser'], options['latencies'],
                  filenames, view.find('dock_area'))

    view.show()
    app.start()
    app.stop()



"""
Collection of classes for handling common data types
"""
from enum import Enum
import functools
import operator

import numpy as np
import pandas as pd
from scipy import signal

from atom.api import Atom, Int, Typed, Value

from .peakdetect import (generate_latencies_bound, generate_latencies_skewnorm,
                         guess, guess_iter, peak_iterator)


@functools.total_ordering
class Point(Enum):

    PEAK = 'peak'
    VALLEY = 'valley'

    def __lt__(self, other):
        if self.__class__ is other.__class__:
            return self.value < other.value
        return NotImplementedError


class ABRWaveform:

    def __init__(self, fs, signal, level):
        self.fs = fs
        self.signal = signal
        self.level = level
        self.points = {}
        self.series = None

    @property
    def x(self):
        return self.signal.index.values

    @property
    def y(self):
        return signal.detrend(self.signal.values)

    def is_subthreshold(self):
        if self.series.threshold is None:
            return False
        return self.level < self.series.threshold

    def is_suprathreshold(self):
        if self.series.threshold is None:
            return True
        return self.level >= self.series.threshold

    def stat(self, lb, ub, func):
        return func(self.signal.loc[lb:ub])

    def mean(self, lb, ub):
        return self.stat(lb, ub, np.mean)

    def std(self, lb, ub):
        return self.stat(lb, ub, np.std)

    def set_point(self, wave, ptype, index):
        if (wave, ptype) not in self.points:
            point = WaveformPoint(self, 0, wave, ptype)
            self.points[wave, ptype] = point
        self.points[wave, ptype].index = int(index)

    def clear_points(self):
        self.points = {}

    def clear_peaks(self):
        for wave, ptype in list(self.points):
            if ptype == Point.PEAK:
                del self.points[wave, ptype]

    def clear_valleys(self):
        for wave, ptype in list(self.points):
            if ptype == Point.VALLEY:
                del self.points[wave, ptype]

    def _set_points(self, guesses, ptype):
        for wave, wave_guess in guesses.iterrows():
            index = wave_guess.get('index', np.nan)
            if not np.isfinite(index):
                index = np.searchsorted(self.x , wave_guess['x'])
                index = np.clip(index, 0, len(self.x)-1)
            else:
                index = int(index)
            self.set_point(wave, ptype, index)


class WaveformPoint(Atom):
    '''
    Parameters
    ----------
    TODO
    '''
    parent = Typed(ABRWaveform)
    index = Int()
    wave_number = Int()
    point_type = Typed(Point)
    iterator = Value()

    def __init__(self, parent, index, wave_number, point_type):
        # Order of setting attributes is important here
        self.parent = parent
        self.point_type = point_type
        self.wave_number = wave_number
        invert = self.is_valley()
        iterator = peak_iterator(parent, index, invert=invert)
        next(iterator)
        self.iterator = iterator
        self.index = index

    def _observe_index(self, event):
        if event['type'] == 'update':
            self.iterator.send(('set', event['value']))

    @property
    def x(self):
        return self.parent.x[self.index]

    @property
    def y(self):
        return self.parent.y[self.index]

    def is_peak(self):
        return self.point_type == Point.PEAK

    def is_valley(self):
        return self.point_type == Point.VALLEY

    @property
    def latency(self):
        latency = self.x
        if self.parent.is_subthreshold():
            return -np.abs(latency)
        return latency

    @property
    def amplitude(self):
        return self.parent.signal.iloc[self.index]

    def move(self, step):
        self.index = self.iterator.send(step)

    def time_to_index(self, time):
        return np.searchsorted(self.parent.x, time)


class ABRSeries(object):

    def __init__(self, waveforms, freq=None, threshold=np.nan):
        waveforms.sort(key=operator.attrgetter('level'))
        self.waveforms = waveforms
        self.freq = freq
        self.threshold = threshold
        for waveform in self.waveforms:
            waveform.series = self

    def get_level(self, level):
        for waveform in self.waveforms:
            if waveform.level == level:
                return waveform
        raise AttributeError(f'{level} dB SPL not in series')

    def guess_p(self, latencies):
        level_guesses = guess_iter(self.waveforms, latencies)
        self._set_points(level_guesses, Point.PEAK)

    def guess_n(self):
        n_latencies = {}
        for w in self.waveforms:
            g = {p.wave_number: p.x for p in w.points.values() if p.is_peak()}
            g = pd.DataFrame({'x': g})
            n_latencies[w.level] = generate_latencies_bound(g)
        level_guesses = guess(self.waveforms, n_latencies, invert=True)
        self._set_points(level_guesses, Point.VALLEY)

    def update_guess(self, level, point):
        waveform = self.get_level(level)
        p = waveform.points[point]
        g = {p.wave_number: p.x}
        g = pd.DataFrame({'x': g})
        latencies = generate_latencies_skewnorm(g)

        i = self.waveforms.index(waveform)
        waveforms = self.waveforms[:i]
        level_guesses = guess_iter(waveforms, latencies, invert=p.is_valley())
        self._set_points(level_guesses, p.point_type)

    def clear_points(self):
        for waveform in self.waveforms:
            waveform.clear_points()

    def clear_peaks(self):
        for waveform in self.waveforms:
            waveform.clear_peaks()

    def clear_valleys(self):
        for waveform in self.waveforms:
            waveform.clear_valleys()

    def _set_points(self, level_guesses, ptype):
        for level, level_guess in level_guesses.items():
            waveform = self.get_level(level)
            waveform._set_points(level_guess, ptype)



import glob
import os.path

import tables
import numpy as np

from abr.datatype import ABRWaveform, ABRSeries


def load(fname, filter=None, abr_window=8.5e-3, frequencies=None):
    with tables.open_file(fname) as fh:
        fs = fh.root.waveforms._v_attrs['fs']
        cutoff = int(abr_window*fs)
        signal = fh.root.waveforms[:, :, :cutoff]*1e6
        levels = fh.root.trial_log.read(field='level')
        available_frequencies = fh.root.trial_log.read(field='frequency')

        # Load all frequencies by default
        if frequencies is None:
            frequencies = np.unique(available_frequencies)

        series = []
        for frequency in frequencies:
            mask = available_frequencies == frequency
            waveforms = [ABRWaveform(fs, s, l, filter=filter, min_latency=2)
                         for s, l in zip(signal[mask], levels[mask])]
            s = ABRSeries(waveforms, frequency/1e3)
            s.filename = fname
            series.append(s)
        return series


def get_frequencies(filename):
    with tables.open_file(filename) as fh:
        frequencies = fh.root.trial_log.read(field='frequency')
        return np.unique(frequencies)


def is_processed(filename, frequency, options):
    from abr.parsers import registry
    filename = registry.get_save_filename(filename, frequency, options)
    return os.path.exists(filename)


def find_unprocessed(dirname, options):
    wildcard = os.path.join(dirname, '*ABR*.hdf5')
    unprocessed = []
    for filename in glob.glob(wildcard):
        for frequency in get_frequencies(filename):
            if not is_processed(filename, frequency*1e-3, options):
                unprocessed.append((filename, frequency))
    return unprocessed



from __future__ import division

import pandas as pd
import numpy as np

from abr.datatype import ABRWaveform, ABRSeries


################################################################################
# Utility functions
################################################################################
def _parse_line(line):
    '''
    Parse list of comma-separated values from line

    Parameters
    ----------
    line : string
        Line containing the values that need to be parsed

    Returns
    -------
    tokens : list
        List of values found in line.  If values are numeric, they will be
        converted to floats.  Otherwise they will be returned as strings.
    '''
    tokens = line.strip().split(',')[1:]
    try:
        return [float(t) for t in tokens if t]
    except ValueError:
        return [t for t in tokens if t]


def load_metadata(filename):
    '''
    Load the metadata stored in the ABR file

    Parameters:
    -----------
    filename : string
        Filename to load

    Returns
    -------
    info : pandas.DataFrame
        Dataframe containing information on each waveform
    '''
    info = {}
    with open(filename, 'r') as fh:
        for i, line in enumerate(fh):
            if i == 20:
                break
            name = line.split(',', 1)[0].strip(':').lower()
            info[name] = _parse_line(line)
    info = pd.DataFrame(info)

    # Number the trials.  We will use this number later to look up which column
    # contains the ABR waveform for corresponding parameter.
    info['waveform'] = np.arange(len(info))
    info.set_index('waveform', inplace=True)

    # Convert the intensity to the actual level in dB SPL
    info['level'] = np.round(info.intensity/10)*10

    # Store the scaling factor for the waveform so we can recover this when
    # loading.  By default the scaling factor is 674. For 110 dB SPL, the
    # scaling factor is 337.  The statistician uses 6.74 and 3.37, but he
    # includes a division of 100 elsewhere in his code to correct.
    info['waveform_sf'] = 6.74e2

    # The rows where level is 110 dB SPL have a different scaling factor.
    info.loc[info.level == 110, 'waveform_sf'] = 3.37e2

    # Start time of stimulus in usec (since sampling period is reported in usec,
    # we should try to be consistent with all time units).
    info['stimulus_start'] = 12.5e3
    return info


def load_waveforms(filename, info):
    '''
    Load the waveforms stored in the ABR file

    Only the waveforms specified in info will be loaded.  For example, if you
    have filtered the info DataFrame to only contain waveforms from channel 1,
    only those waveforms will be loaded.

    Parameters:
    -----------
    filename : string
        Filename to load
    info : pandas.DataFrame
        Waveform metadata (see `load_metadata`)

    Returns
    -------
    info : pandas.DataFrame
        Dataframe containing waveforms

    '''
    # Read the waveform table into a dataframe
    df = pd.io.parsers.read_csv(filename, skiprows=20)

    # Keep only the columns containing the signal of interest.  There are six
    # columns for each trial.  We only want the column containing the raw
    # average (i.e., not converted to uV).
    df = df[[c for c in df.columns if c.startswith('Average:')]]

    # Renumber them so we can look them up by number.  The numbers should
    # correspond to the trial number we generated in `load_metadata`.
    df.columns = np.arange(len(df.columns))

    # Loop through the entries in the info DataFrame.  This dataframe contains
    # metadata needed for processing the waveform (e.g., it tells us which
    # waveforms to keep, the scaling factor to use, etc.).
    signals = []
    for w_index, w_info in info.iterrows():
        # Compute time of each point.  Currently in usec because smp. period is
        # in usec.
        t = np.arange(len(df), dtype=np.float32)*w_info['smp. period']
        # Subtract stimulus start so that t=0 is when stimulus begins.  Convert
        # to msec.
        t = (t-w_info['stimulus_start'])*1e-3
        time = pd.Index(t, name='time')

        # Divide by the scaling factor and convert from nV to uV
        s = df[w_index]/w_info['waveform_sf']*1e-3
        s.index = time
        signals.append(s)

    # Merge together the waveforms into a single DataFrame
    waveforms = pd.concat(signals, keys=info.index, names=['waveform'])
    waveforms = waveforms.unstack(level='waveform')
    return waveforms


################################################################################
# API
################################################################################

# Minimum wave 1 latencies
latencies = {
    1000: 3.1,
    3000: 2.1,
    4000: 2.3,
    6000: 1.8,
}

def load(fname, filter=None, abr_window=8.5e-3):
    with open(fname) as fh:
        line = fh.readline()
        if not line.startswith('Identifier:'):
            raise IOError('Unsupported file format')
    info = load_metadata(fname)
    info = info[info.channel == 1]
    fs = 1/(info.iloc[0]['smp. period']*1e-6)
    series = []
    for frequency, f_info in info.groupby('stim. freq.'):
        signal = load_waveforms(fname, f_info)
        signal = signal[signal.index >= 0]
        waveforms = []
        min_latency = latencies.get(frequency)
        for i, row in f_info.iterrows():
            s = signal[i].values[np.newaxis]
            waveform = ABRWaveform(fs, s, row['level'], min_latency=min_latency,
                                   filter=filter)
            waveforms.append(waveform)

        s = ABRSeries(waveforms, frequency/1e3)
        s.filename = fname
        series.append(s)
    return series




import enaml
from enaml.qt.qt_application import QtApplication

with enaml.imports():
    from abr.launch_window import LaunchWindow


def main():
    app = QtApplication()
    window = LaunchWindow()
    window.show()
    app.start()
    app.stop()



'''
This module defines the import/export routines for interacting with the data
store.  If you wish to customize this, simply define the load() function.

load(run_location, invert, filter) -- When the program needs a run loaded, it
will pass the run_location provided by list().  Invert is a boolean flag
indicating whether waveform polarity should be flipped.  Filter is a dictionary
containing the following keys:
    1. ftype: any of None, butterworth, bessel, etc.
    2. fh: highpass cutoff (integer in Hz)
    3. fl: lowpass cutoff (integer in Hz)
All objects of the epl.datatype.Waveform class will accept the filter
dictionary and perform the appropriate filtering.  It is recommended you use the
filtering provided by the Waveform class as the parameters of the filter will
also be recorded.  This function must return an object of the
epl.datatype.ABRSeries class.  See this class for appropriate documentation.

The save function must return a message.  If there is an error in saving, throw
the appropriate exception.
'''

import importlib
import re
from glob import glob
import os
from pathlib import Path
import time

import pandas as pd
import numpy as np

from ..datatype import Point

P_ANALYZER = re.compile('.*kHz(?:-(\w+))?-analyzed.txt')


def get_analyzer(filename):
    return P_ANALYZER.match(filename.name).group(1)


def waveform_string(waveform):
    data = [f'{waveform.level:.2f}']
    data.append(f'{waveform.mean(0, 1)}')
    data.append(f'{waveform.std(0, 1)}')
    for _, point in sorted(waveform.points.items()):
        data.append(f'{point.latency:.8f}')
        data.append(f'{point.amplitude:.8f}')
    return '\t'.join(data)


def filter_string(waveform):
    if getattr(waveform, '_zpk', None) is None:
        return 'No filtering'
    t = 'Pass %d -- z: %r, p: %r, k: %r'
    filt = [t % (i, z, p, k) for i, (z, p, k) in enumerate(waveform._zpk)]
    return '\n' + '\n'.join(filt)


def load_analysis(fname):
    th_match = re.compile('Threshold \(dB SPL\): ([\w.]+)')
    freq_match = re.compile('Frequency \(kHz\): ([\d.]+)')
    with open(fname) as fh:
        text = fh.readline()
        th = th_match.search(text).group(1)
        th = None if th == 'None' else float(th)
        text = fh.readline()
        freq = float(freq_match.search(text).group(1))

        for line in fh:
            if line.startswith('NOTE'):
                break
        data = pd.io.parsers.read_csv(fh, sep='\t', index_col='Level')
    return (freq, th, data)


def parse_peaks(peaks, threshold):
    # Convert the peaks dataframe to a format that can be used by _set_points.
    p_pattern = re.compile('P(\d) Latency')
    n_pattern = re.compile('N(\d) Latency')

    p_latencies = {}
    n_latencies = {}

    for c in peaks:
        match = p_pattern.match(c)
        if match:
            wave = int(match.group(1))
            p_latencies[wave] = pd.DataFrame({'x': peaks[c]})
        match = n_pattern.match(c)
        if match:
            wave = int(match.group(1))
            n_latencies[wave] = pd.DataFrame({'x': peaks[c]})

    p_latencies = pd.concat(p_latencies.values(), keys=p_latencies.keys(),
                            names=['wave'])
    p_latencies = {g: df.reset_index('Level', drop=True) \
                   for g, df in p_latencies.groupby('Level')}
    n_latencies = pd.concat(n_latencies.values(), keys=n_latencies.keys(),
                            names=['wave'])
    n_latencies = {g: df.reset_index('Level', drop=True) \
                   for g, df in n_latencies.groupby('Level')}

    for level, df in p_latencies.items():
        if level < threshold:
            df[:] = -df[:]

    for level, df in n_latencies.items():
        if level < threshold:
            df[:] = -df[:]

    return p_latencies, n_latencies


class Parser(object):

    filename_template = '{filename}-{frequency}kHz-{user}analyzed.txt'

    def __init__(self, file_format, filter_settings, user=None):
        '''
        Parameters
        ----------
        file_format : string
            File format that will be loaded.
        filter_settings : {None, dict}
            If None, no filtering is applied. If dict, must contain ftype,
            lowpass, highpass and order as keys.
        user : {None, string}
            Person analyzing the data.
        '''
        self._file_format = file_format
        self._filter_settings = filter_settings
        self._user = user
        self._module_name = f'abr.parsers.{file_format}'
        self._module = importlib.import_module(self._module_name)

    def load(self, filename, frequencies=None):
        return self._module.load(filename, self._filter_settings, frequencies)

    def load_analysis(self, series, filename):
        freq, th, peaks = load_analysis(filename)
        series.threshold = th
        p_latencies, n_latencies = parse_peaks(peaks, th)
        series._set_points(p_latencies, Point.PEAK)
        series._set_points(n_latencies, Point.VALLEY)

    def find_analyzed_files(self, filename, frequency):
        frequency = round(frequency * 1e-3, 8)
        glob_pattern = self.filename_template.format(
            filename=filename.with_suffix(''),
            frequency=frequency,
            user='*')
        path = Path(glob_pattern)
        return list(path.parent.glob(path.name))

    def get_save_filename(self, filename, frequency):
        # Round frequency to nearest 8 places to minimize floating-point
        # errors.
        user_name = self._user + '-' if self._user else ''
        frequency = round(frequency * 1e-3, 8)
        save_filename = self.filename_template.format(
            filename=filename.with_suffix(''),
            frequency=frequency,
            user=user_name)
        return Path(save_filename)

    def save(self, model):
        # Assume that all waveforms were filtered identically
        filter_history = filter_string(model.waveforms[-1])

        # Generate list of columns
        columns = ['Level', '1msec Avg', '1msec StDev']
        point_keys = sorted(model.waveforms[0].points)
        for point_number, point_type in point_keys:
            point_type_code = 'P' if point_type == Point.PEAK else 'N'
            for measure in ('Latency', 'Amplitude'):
                columns.append(f'{point_type_code}{point_number} {measure}')

        columns = '\t'.join(columns)
        spreadsheet = '\n'.join(waveform_string(w) \
                                for w in reversed(model.waveforms))
        content = CONTENT.format(threshold=model.threshold,
                                 frequency=model.freq*1e-3,
                                 filter_history=filter_history,
                                 columns=columns,
                                 spreadsheet=spreadsheet)

        filename = self.get_save_filename(model.filename, model.freq)
        with open(filename, 'w') as fh:
            fh.writelines(content)

    def find_all(self, dirname, frequencies=None):
        result = self._module.find_all(dirname, self._filter_settings)
        if frequencies is not None:
            if np.isscalar(frequencies):
                frequencies = [frequencies]
            result = [(p, f) for (p, f) in result if f in frequencies]
        return result

    def find_processed(self, dirname, frequencies=None):
        return [(p, f) for p, f in self.find_all(dirname, frequencies) \
                if self.get_save_filename(p, f).exists()]

    def find_unprocessed(self, dirname, frequencies=None):
        return [(p, f) for p, f in self.find_all(dirname, frequencies) \
                if not self.get_save_filename(p, f).exists()]

    def find_analyses(self, dirname, frequencies=None):
        analyzed = {}
        for p, f in self.find_all(dirname, frequencies):
            analyzed[p, f] = self.find_analyzed_files(p, f)
        return analyzed

    def load_analyses(self, dirname, frequencies=None):
        analyzed = self.find_analyses(dirname, frequencies)
        keys = []
        thresholds = []
        for (raw_file, frequency), analyzed_files in analyzed.items():
            for analyzed_file in analyzed_files:
                user = get_analyzer(analyzed_file)
                keys.append((raw_file, frequency, analyzed_file, user))
                _, threshold, _ = load_analysis(analyzed_file)
                thresholds.append(threshold)

        cols = ['raw_file', 'frequency', 'analyzed_file', 'user']
        index = pd.MultiIndex.from_tuples(keys, names=cols)
        return pd.Series(thresholds, index=index)


CONTENT = '''
Threshold (dB SPL): {threshold:.2f}
Frequency (kHz): {frequency:.2f}
Filter history (zpk format): {filter_history}
NOTE: Negative latencies indicate no peak
{columns}
{spreadsheet}
'''.strip()


PARSER_MAP = {
    'PSI': 'psiexperiment',
    'EPL': 'EPL CFTS',
    'NCRAR': 'IHS text export',
}



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import sys

from setuptools import find_packages, setup

with open(os.path.join(os.path.dirname(__file__), 'README.md')) as f:
    readme = f.read()

packages = [
    'abr',
]

package_data = {
    'abr': ['*.enaml', 'data/*'],
}

requires = [
    'atom',
    'enaml',
    'numpy',
    'matplotlib',
    'scipy',
    'pandas',
]

classifiers = [
    'Operating System :: OS Independent',
    'Programming Language :: Python :: 3',
    'License :: OSI Approved :: BSD License',
]

setup(
    name='ABR',
    version='0.0.1',
    author='Brad Buran',
    author_email='bburan@alum.mit.edu',
    description='ABR wave analyzer',
    long_description=readme,
    long_description_content_type='text/markdown',
    url='https://github.com/bburan/abr',
    packages=find_packages(),
    package_data=package_data,
    install_requires=requires,
    classifiers=classifiers,
    entry_points={
        'console_scripts': [
            'abr = abr.app.cmd_launcher:main',
            'abr_gui = abr.app.cmd_main:main',
            'abr_loop = abr.app.cmd_main_loop:main',
        ]
    },
)



from abr.parsers import Parser

from scipy import stats


P_LATENCIES = {
    1: stats.norm(1.5, 0.5),
    2: stats.norm(2.5, 1),
    3: stats.norm(3.0, 1),
    4: stats.norm(4.0, 1),
    5: stats.norm(5.0, 2),
}


def add_default_arguments(parser, waves=True):
    parser.add_argument('--nofilter', action='store_false', dest='filter',
                        default=True, help='Do not filter waveform')
    parser.add_argument('--lowpass',
                        help='Lowpass cutoff (Hz), default 3000 Hz',
                        default=3000, type=float)
    parser.add_argument('--highpass',
                        help='Highpass cutoff (Hz), default 300 Hz',
                        default=300, type=float)
    parser.add_argument('--order',
                        help='Filter order, default 1st order', default=1,
                        type=int)
    parser.add_argument('--parser', default='HDF5', help='Parser to use')
    parser.add_argument('--user', help='Name of person analyzing data')

    if waves:
        group = parser.add_mutually_exclusive_group(required=True)
        group.add_argument('--threshold-only', action='store_true')
        group.add_argument('--all-waves', action='store_true')
        group.add_argument('--waves', type=int, nargs='+')

def parse_args(parser, waves=True):
    options = parser.parse_args()
    exclude = ('filter', 'lowpass', 'highpass', 'order', 'parser', 'user',
               'waves', 'all_waves', 'threshold_only')
    new_options = {k: v for k, v in vars(options).items() if k not in exclude}
    filter_settings = None
    if options.filter:
        filter_settings = {
            'lowpass': options.lowpass,
            'highpass': options.highpass,
            'order': options.order,
        }
    new_options['parser'] = Parser(options.parser, filter_settings,
                                   options.user)

    if not waves:
        return new_options

    if options.all_waves:
        waves = [1, 2, 3, 4, 5]
    elif options.threshold_only:
        waves = []
    else:
        waves = options.waves[:]
    new_options['latencies'] = {w: P_LATENCIES[w] for w in waves}
    return new_options
